/*
 * exception.c
 * Handle exceptions for the translator.
 *
 * Copyright (c) 1996 Systems Architecture Research Centre,
 *		   City University, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@sarc.city.ac.uk>, June 1996.
 */

#include "config.h"
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <signal.h>
#include "gtypes.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "classMethod.h"
#include "exception.h"
#include "baseClasses.h"
#include "lookup.h"
#include "thread.h"
#include "md.h"

object*	ClassFormatError;
object*	LinkageError;
object*	ClassNotFoundException;
object*	NoSuchFieldError;
object*	NoSuchMethodError;
object*	OutOfMemoryError;
object*	UnsatisfiedLinkError;
object*	VirtualMachineError;
object*	ClassCircularityError;
object*	NegativeArraySizeException;
object*	ClassCastException;
object*	IllegalMonitorStateException;
object*	NullPointerException;
object*	ArithmeticException;
object*	ArrayIndexOutOfBoundsException;

extern object* exceptionObject;

#if defined(__WIN32__)
#define	SIG_T	void(*)()
#else
#define	SIG_T	void*
#endif

static void dispatchException(object*, exceptionFrame*);
static void nullException(EXCEPTIONPROTO);
static void arithmeticException(EXCEPTIONPROTO);

/*
 * Setup the internal exceptions.
 */
void
initExceptions(void)
{
#define	EXCEPTION(s) alloc_object(lookupClass(addString(#s)), true)

	ClassFormatError = EXCEPTION(java/lang/ClassFormatError);
	LinkageError = EXCEPTION(java/lang/LinkageError);
	ClassNotFoundException = EXCEPTION(java/lang/ClassNotFoundException);
	NoSuchFieldError = EXCEPTION(java/lang/NoSuchFieldError);
	NoSuchMethodError = EXCEPTION(java/lang/NoSuchMethodError);
	OutOfMemoryError = EXCEPTION(java/lang/OutOfMemoryError);
	UnsatisfiedLinkError = EXCEPTION(java/lang/UnsatisfiedLinkError);
	VirtualMachineError = EXCEPTION(java/lang/VirtualMachineError);
	ClassCircularityError = EXCEPTION(java/lang/ClassCircularityError);
	NegativeArraySizeException = EXCEPTION(java/lang/NegativeArraySizeException);
	ClassCastException = EXCEPTION(java/lang/ClassCastException);
	IllegalMonitorStateException = EXCEPTION(java/lang/IllegalMonitorStateException);
	NullPointerException = EXCEPTION(java/lang/NullPointerException);
	ArithmeticException = EXCEPTION(java/lang/ArithmeticException);
	ArrayIndexOutOfBoundsException = EXCEPTION(java/lang/ArrayIndexOutOfBoundsException);

	initClasses();

#if !defined(DEBUG)
	/* Catch signal we need to convert to exceptions */
#if defined(SIGSEGV)
	signal(SIGSEGV, (SIG_T)nullException);
#endif
#if defined(SIGBUS)
	signal(SIGBUS, (SIG_T)nullException);
#endif
#if defined(SIGFPE)
	signal(SIGFPE, (SIG_T)arithmeticException);
#endif
#if defined(SIGPIPE)
	signal(SIGPIPE, SIG_IGN);
#endif
#endif
}

/*
 * Throw an exception.
 */
void
throwException(object* eobj)
{
	exceptionFrame frame;

	if (eobj == 0) {
		fprintf(stderr, "Exception thrown on null object ... aborting\n");
		abort();
		exit(1);
	}
	FIRSTFRAME(frame, eobj);
	dispatchException(eobj, &frame);
}

/*
 * Search for a matching exception.
 */
static
void
dispatchException(object* eobj, exceptionFrame* frame)
{
	exceptionInfo einfo;
	classes* class;

	class = eobj->mtable->class;

	for (; FRAMEOKAY(frame); frame = NEXTFRAME(frame)) {
		findExceptionInMethod(PCFRAME(frame), class, &einfo);
		if (einfo.handler != 0) {
			CALL_KAFFE_EXCEPTION(frame, einfo, eobj);
		}
	}

	fprintf(stderr, "Failed to catch exception ... aborting:\n\t%s\n", eobj->mtable->class->name);
	fflush(stderr);
	exit(1);
}

/*
 * Null exception - catches bad memory accesses.
 */
static
void
nullException(EXCEPTIONPROTO)
{
	exceptionFrame frame;

#if !defined(DEBUG)
	/* Reset signal handler - necessary for SysV, does no harm for BSD */
	signal(sig, (SIG_T)nullException);
#endif
	EXCEPTIONFRAME(frame, ctx);
	dispatchException(NullPointerException, &frame);
}

/*
 * Division by zero.
 */
static
void
arithmeticException(EXCEPTIONPROTO)
{
	exceptionFrame frame;

#if !defined(DEBUG)
        /* Reset signal handler - necessary for SysV, does no harm for BSD */
        signal(sig, (SIG_T)arithmeticException);
#endif
	EXCEPTIONFRAME(frame, ctx);
	dispatchException(ArithmeticException, &frame);
}
